/*
 * host.h - host header file
 *
 * Copyright (C) 2021 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: nevadita.chatterjee@iopsys.eu
 *
 */
#ifndef HOSTD_H
#define HOSTD_H

#include <easy/easy.h>
#include "topologyd.h"

#define MAP_TLV_CLIENT_ASSOCIATION_EVENT 0x92
#define MAP_TLV_ASSOCIATED_CLIENTS      0x84

#define HOST_NODE_MAX    500
#define HOST_MAX_IPV6ADDR 20
#define HOST_MAX_IPV4ADDR 20
#define HOST_REFRESH_TIMER 15
#define NEIGH_CMD "ip neigh show nud reachable|awk '{print $5}'|grep %s"
#define ARPING_CMD "arping -I br-lan -c 1 -w 1 -D %s &"
#define HOST_ARP_TIMER 2
#define HOST_WIFI_TIMER 5
#define WIFI_TIME_OUT 36
#define WIFI_EVENTS	"wifi.sta"
//The below timer represent ageout as 24hrs
#define HOST_AGE_TIMER_DEFAULT 1440

#define IOPSYS_OUI_1 0x00
#define IOPSYS_OUI_2 0x22
#define IOPSYS_OUI_3 0x07

#define MIN_VENDOR_TLV_LEN 5
#define HW_ADDR_LENGTH 6

enum host_types {
	HOST_TYPE_ETHER = 0,
	HOST_TYPE_GHN,
	HOST_TYPE_HPNA,
	HOST_TYPE_HOMEPLUG,
	HOST_TYPE_MOCA,
	HOST_TYPE_UPA,
	HOST_TYPE_WIFI,
	HOST_TYPE_OTHER
};

enum address_type {
	HOST_ADDR_STATIC = 0,
	HOST_ADDR_DHCP,
};

struct host_topo_query {
	struct topologyd_private *p;
	uint8_t dst_mac[6];
};

struct host_node {
	uint8_t hwaddr[6];       //macaddress of the node
	uint8_t link_hwaddr[6];       //macaddress of the link interface
	char device[16];
	char parent_device[16];
	char network[16];
	struct hlist_node hlist;
	uint8_t active;
	time_t active_last_change;
	uint8_t is_ipaddr; /*param to check if the ip address is filled*/
	char ipv4addr[24];
	char hostname[64];
	uint32_t ipv6addr_count;
	char ipv6addr[HOST_MAX_IPV6ADDR][128];
	uint32_t ipv4addr_count;
	char ipv4addr_list[HOST_MAX_IPV4ADDR][24];
	enum host_types intf_type;
	uint8_t is_ieee1905_node;
	bool is_in_assoclist;
	bool ignore;
	bool is_copy; // This variable is used to keep track that the node is filled by copy
	bool is_event; // This variable is used to keep track that the node is filled by copy and event needs to be sent
	struct {
		unsigned long long rx_bytes;
		unsigned long long tx_bytes;
		unsigned long long rx_packets;
		unsigned long long tx_packets;
	} stats;
	struct hlist_head nf_tc;
	//char addrsrc[8];
	enum address_type addrsrc;
	unsigned int leasetmrmn;
	int frequency;
	int rssi;
	int snr;
	int idle;
	int in_network;
	int port_no;
	int active_connections;
	struct host_node *al_node;
	//struct timeval age_time; // This time is used to keep track of wifi hosts
	time_t age_time;
	bool is_copy_lan;
	bool is_device_mac;
};

int host_config_defaults(struct host_config *cfg);
void host_remove_aged_node(struct topologyd_private *priv);
void host_event_handler(struct ubus_context *ctx,
		struct ubus_event_handler *ev, const char *type,
		struct blob_attr *msg);
int host_nodes(struct ubus_context *ctx,
		struct ubus_object *obj,
		struct ubus_request_data *req,
		const char *method,
		struct blob_attr *msg);
void host_dump_node_nbr(struct topologyd_private *priv,
		struct blob_buf *b);
struct host_node *host_topo_node_add(struct topologyd_private *priv,
		struct node *node, uint8_t *mac_addr, uint8_t is_wifi);
struct host_node *host_topo_node_del(struct topologyd_private *priv,
		struct node *node, uint8_t *mac_addr, uint8_t is_wifi);
int host_get_ipaddr_hostname(struct topologyd_private *priv,
		char *macaddr_str, struct host_node *p);
void host_append_dhcpv6_info(struct host_node *p);
int host_append_dhcpv4_info(char *mac_addr, struct host_node *p);
int host_add_netlink(int *fd);
void host_netlink_cb(struct uloop_fd *fd, unsigned int events);
void host_run_status_check(struct topologyd_private *priv);
void host_periodic_refresh(struct uloop_timeout *t);
void host_process_topo_notification(uint8_t *tlv, struct topologyd_private *priv);
void host_agent_event_handler(struct ubus_context *ctx,
		struct ubus_event_handler *ev, const char *type,
		struct blob_attr *msg);
void host_set_interface_type(struct host_node *p, struct topologyd_private *priv);
void host_arp_table_refresh(struct uloop_timeout *t);
void host_wifi_table_refresh(struct uloop_timeout *t);
void host_get_network_dump(struct topologyd_private *priv);
int topologyd_send_host_topology_query(struct topologyd_private *p);
int host_send_mid_topology_query(struct topologyd_private *p, uint8_t *dest_mac);

bool host_is_in_assoclist(struct topologyd_private *priv, uint8_t *hwaddr);

void host_change_active_state(struct topologyd_private *priv, struct host_node *p, uint8_t state);

int host_copy_node_data(struct host_node *src_host,  struct host_node *dest_host,
		 struct topologyd_private *priv);

void host_send_client_event(struct topologyd_private *priv, struct host_node *p, uint8_t state);


bool host_create_file(void);
bool is_local_mapagent_available(void);

int host_get_wifi_data(struct topologyd_private *priv, struct host_node *p);
void host_send_topology_query_all(struct topologyd_private *priv);
int host_node_get_statistics(struct topologyd_private *priv);
int host_is_device_address(struct topologyd_private *priv, char *address, int family);
int host_get_interface_type(struct topologyd_private *priv);
int host_send_arp_req(char *ipaddr_str);

#endif /* HOSTD_H */
