
#include <map2.h>
#include <1905_tlvs.h>
#include <cmdu.h>

#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>

#include <libubus.h>
#include <uci.h>
#include <easy/easy.h>

#include "debug.h"
#include "host_config.h"
#include "host.h"
#include "cmdu_validate.h"
#include "ieee1905_deps.h"
#include "host_ieee1905_utils.h"

static int check_serialized_tlv(struct tlv *t, uint16_t len)
{
	uint8_t *tv_data;
	uint16_t tlv_len;

	if (!t)
		return -1;

	tlv_len = t_tlv_length(t);

	if (tlv_len != len)
		return -1;

	tv_data = (uint8_t *)t->data;
	if (!tv_data)
		return -1;

	return 0;
}

/* Check SupportedFreqBand TLV */
static int check_supported_band_tlv(struct tlv *t)
{
	/* band (1 byte) */
	return check_serialized_tlv(t,
				sizeof(struct tlv_supported_band));
}

/* Check SupportedRole TLV */
static int check_supported_role_tlv(struct tlv *t)
{
	/* role (1 byte) */
	return check_serialized_tlv(t,
				sizeof(struct tlv_supported_role));
}

/* Check MacAddress TLV */
static int check_mac_addr_tlv(struct tlv *t)
{
	/* mac (6 byte) */
	return check_serialized_tlv(t,
			sizeof(struct tlv_macaddr));
}

/* Check ALMacAddress TLV */
static int check_almac_addr_tlv(struct tlv *t)
{
	/* almac (6 byte) */
	return check_serialized_tlv(t,
			sizeof(struct tlv_aladdr));
}

/* Check ProfileVersion TLV */
static int check_profile_ver_tlv(struct tlv *t)
{
	/* version (1 byte) */
	return check_serialized_tlv(t,
			sizeof(struct tlv_1905_profile));
}

/* Check DeviceIdentification TLV */
static int check_device_id_tlv(struct tlv *t)
{
	/* device info (63*3 byte) */
	return check_serialized_tlv(t,
			sizeof(struct tlv_device_identification));
}

bool validate_topology_discovery(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = TLV_TYPE_AL_MAC_ADDRESS_TYPE,
			.present = TLV_PRESENT_ONE,
			.minlen = 6,
			.maxlen = 6,
		},
		[1] = {
			.type = TLV_TYPE_MAC_ADDRESS_TYPE,
			.present = TLV_PRESENT_ONE,
			.minlen = 6,
			.maxlen = 6,
		}
	};

	dbg("%s: --->\n", __func__);

	ret = t_cmdu_parse_tlvs(cmdu, tv, d_policy, 2);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	/* TLV_TYPE_AL_MAC_ADDRESS_TYPE */
	if (check_almac_addr_tlv(tv[0][0]))
		return false;

	/* TLV_TYPE_MAC_ADDRESS_TYPE */
	if (check_mac_addr_tlv(tv[1][0]))
		return false;

	return true;
}


bool validate_topology_response(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	dbg("%s: --->\n", __func__);

	int ret = 0;
	struct tlv_policy a_policy[] = {
		[0] = { .type = TLV_TYPE_DEVICE_INFORMATION_TYPE,
				.present = TLV_PRESENT_ONE,
				.minlen = 7,
		},
		[1] = { .type = TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES,
				.present = TLV_PRESENT_OPTIONAL_MORE,
				.minlen = 1,
		},
		[2] = { .type = TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST,
				.present = TLV_PRESENT_OPTIONAL_MORE,
				.minlen = 6,
		},
		[3] = { .type = TLV_TYPE_NEIGHBOR_DEVICE_LIST,
				.present = TLV_PRESENT_OPTIONAL_MORE,
				.minlen = 6,
		},
		[4] = { .type = TLV_TYPE_POWER_OFF_INTERFACE,
				.present = TLV_PRESENT_OPTIONAL_MORE,
		},
		[5] = { .type = TLV_TYPE_L2_NEIGHBOR_DEVICE,
				.present = TLV_PRESENT_OPTIONAL_MORE,
		},
		[6] = { .type = MAP_TLV_SUPPORTED_SERVICE,
				.present = TLV_PRESENT_OPTIONAL_MORE,
				.minlen = 1, /* num_services(==0) */
		},
		[7] = { .type = MAP_TLV_AP_OPERATIONAL_BSS,
				.present = TLV_PRESENT_OPTIONAL_MORE,
				.minlen = 1, /* tlv_ap_oper_bss:num_radio */
		},
		[8] = { .type = MAP_TLV_ASSOCIATED_CLIENTS,
				.present = TLV_PRESENT_OPTIONAL_MORE,
				.minlen = 1, /* tlv_assoc_client:num_bss */
		},
		[9] = { .type = MAP_TLV_MULTIAP_PROFILE,
				.present = TLV_PRESENT_OPTIONAL_MORE,
				.minlen = 1, /* tlv_map_profile:profile */
				.maxlen = 1,
		},
		[10] = { .type = TLV_TYPE_VENDOR_SPECIFIC,
				 .present = TLV_PRESENT_OPTIONAL_MORE,
				 .minlen = 2, /* tlv_vendor_metric_specific:oui+num_sta */
		}
	};
	int num = 0;

	dbg("parsing topology response |" MACFMT "|CMDU: topology response\n",
			MAC2STR(cmdu->origin));

	ret = t_cmdu_parse_tlvs(cmdu, tv, a_policy, 11);
	if (ret) {
		dbg("%s: parse_tlv failed\n", __func__);
		return false;
	}

	/* TLV_TYPE_DEVICE_INFORMATION_TYPE */
	if (tv[0][0]) {
		uint8_t *tv_data;
		struct tlv_device_info *tlv;
		uint16_t tlv_len = 0;
		int i = 0, offset = 0;

		tlv_len = t_tlv_length(tv[0][0]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_device_info *)tv[0][0]->data;
		if (!tlv)
			return false;

		tv_data = (uint8_t *)tlv;

		offset += 7; /*already checked minlen aladdr(6 bytes) + num_interface(1 byte)*/

		for (i = 0; i < tlv->num_interface; i++) {
			uint8_t num_mediainfo = 0;

			/* macaddr (6 bytes) + mediatype (2 byte) + num_mediainfo (1 byte)*/
			if (offset + 9 > tlv_len)
				return false;
			/*macaddr*/
			offset += 6;
			/*mediatype*/
			offset += 2;

			memcpy(&num_mediainfo, &tv_data[offset], 1);
			/*num_mediainfo*/
			offset += 1;

			if (offset + num_mediainfo > tlv_len)
				return false;
			/*num_mediainfo (1 byte)*/
			offset += num_mediainfo;
		}
	}

	/* TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES */
	while (tv[1][num]) {
		uint8_t *tv_data;
		struct tlv_device_bridge_caps *tlv;
		uint16_t tlv_len = 0;
		int i = 0, offset = 0;

		tlv_len = t_tlv_length(tv[1][num]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_device_bridge_caps *)tv[1][num]->data;
		if (!tlv)
			return false;

		tv_data = (uint8_t *)tlv;

		/*num_tuples (1 byte)*/
		if (offset + 1 > tlv_len)
			return false;

		offset += 1;

		for (i = 0; i < tlv->num_tuples; i++) {
			uint8_t num_macaddrs = 0;

			if (offset + 1 > tlv_len)
				return false;

			memcpy(&num_macaddrs, &tv_data[offset], 1);
			/*num_macaddrs (1 byte)*/
			offset += 1;

			if ((offset + (num_macaddrs*6)) > tlv_len)
				return false;
			/*macaddr (6 byte) */
			offset += (num_macaddrs*6);
		}
		num++;
	}

	num = 0;

	/* TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST */
	while (tv[2][num]) {
		struct tlv_non1905_neighbor *tlv;
		uint16_t tlv_len = 0;
		int size = 0;

		tlv_len = t_tlv_length(tv[2][num]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_non1905_neighbor *)tv[2][num]->data;
		if (!tlv)
			return false;

		size = tlv_len - (sizeof(struct tlv_non1905_neighbor));

		if (size < 0)
			return false;
		else if (size > 0) {
			size = size % (sizeof(struct non1905_neighbor));
			if (size)
				return false;
		}
		num++;
	}

	num = 0;
	/* TLV_TYPE_NEIGHBOR_DEVICE_LIST */
	while (tv[3][num]) {
		struct tlv_1905neighbor *tlv;
		uint16_t tlv_len = 0;
		int size = 0;

		tlv_len = t_tlv_length(tv[3][num]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_1905neighbor *)tv[3][num]->data;
		if (!tlv)
			return false;

		size = tlv_len - (sizeof(struct tlv_1905neighbor));

		if (size < 0)
			return false;
		else if (size > 0) {
			size = size % (sizeof(struct i1905_neighbor));
			if (size)
				return false;
		}
		num++;
	}
	num = 0;
	/* MAP_TLV_SUPPORTED_SERVICE */
	if (tv[6][0]) {
		uint8_t *tv_data;
		uint16_t tlv_len = 0;
		uint8_t num_services;
		int offset = 0;

		tlv_len = t_tlv_length(tv[6][0]);
		if (!tlv_len)
			return false;

		tv_data = (uint8_t *) tv[6][0]->data;
		if (!tv_data)
			return false;

		/* num_services (1 byte) */
		if (offset + 1 > tlv_len)
			return false;

		offset += 1;

		num_services = tv_data[0];

		/* services (num_services bytes) */
		if (offset + num_services > tlv_len)
			return false;
	}

	/* MAP_TLV_AP_OPERATIONAL_BSS */
	if (tv[7][0]) {
		struct tlv_ap_oper_bss *tlv;
		uint8_t *tv_data;
		uint16_t tlv_len = 0;
		int i, offset = 0;

		tlv_len = t_tlv_length(tv[7][0]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_ap_oper_bss *)tv[7][0]->data;
		if (!tlv)
			return false;

		tv_data = (uint8_t *)tlv;

		/* num_radio (1 byte) */
		if (offset + 1 > tlv_len)
			return false;

		offset += 1;

		for (i = 0; i < tlv->num_radio; i++) {
			uint8_t num_bss = 0;
			int j;

			/* macaddr (6 bytes) */
			if (offset + 6 > tlv_len)
				return false;

			offset += 6;

			/* num_bss (1 byte) */
			if (offset + 1 > tlv_len)
				return false;

			memcpy(&num_bss, &tv_data[offset], 1);

			offset += 1;

			for (j = 0; j < num_bss; j++) {
				uint8_t ssidlen = 0;

				/* macaddr (6 bytes) */
				if (offset + 6 > tlv_len)
					return false;

				offset += 6;

				/* ssidlen (1 byte) */
				if (offset + 1 > tlv_len)
					return false;

				memcpy(&ssidlen, &tv_data[offset], 1);

				offset += 1;

				/* ssid (ssidlen bytes) */
				if (offset + ssidlen > tlv_len)
					return false;

				offset += ssidlen;
			}
		}
	}

	/* MAP_TLV_ASSOCIATED_CLIENTS */
	if (tv[8][0]) {
		uint16_t tlv_len = 0;
		struct tlv_assoc_client *tlv;
		uint8_t *tv_data;
		int i, offset = 0;

		tlv_len = t_tlv_length(tv[8][0]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_assoc_client *)tv[8][0]->data;
		if (!tlv)
			return false;

		tv_data = (uint8_t *)tlv;

		/* num_bss (1 byte) */
		if (offset + 1 > tlv_len)
			return false;

		offset += 1;

		for (i = 0; i < tlv->num_bss; i++) {
			uint16_t num_client = 0;

			/* radio id (6 bytes) */
			if (offset + 6 > tlv_len)
				return false;

			offset += 6;

			/* num_client (2 bytes) */
			if (offset + 2 > tlv_len)
				return false;

			num_client = BUF_GET_BE16(tv_data[offset]);

			offset += 2;

			 /* num_client * (macaddr + conntime) */
			if (offset + (num_client * 8) > tlv_len)
				return false;

			offset += (num_client * 8);
		}
	}

	/* MAP_TLV_MULTIAP_PROFILE */
	if (tv[9][0]) {
		uint16_t tlv_len = 0;
		struct tlv_map_profile *tlv;

		tlv_len = t_tlv_length(tv[9][0]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_map_profile *)tv[9][0]->data;
		if (!tlv)
			return false;

		if (tlv_len < sizeof(struct tlv_map_profile))
			return false;
	}

	num = 0;
	/* TLV_TYPE_VENDOR_SPECIFIC */
	while (tv[10][num]) {
		uint16_t tlv_len = 0;
		struct tlv_vendor_metric_specific *tlv;
		int i, offset = 0;
		uint8_t *t = (uint8_t *) tv[10][num]->data;
		uint16_t num_sta = 0;

		tlv_len = t_tlv_length(tv[10][num]);

		if (!tlv_len)
			return false;

		tlv = (struct tlv_vendor_metric_specific *)tv[10][num]->data;
		if (!tlv)
			return false;

		/* oui (3 bytes) */
		if (offset + 3 > tlv_len)
			return false;

		/* TODO: move to IOPSYS OUI */
		if (tlv->oui[0] != 0x00 || tlv->oui[1] != 0x22 || tlv->oui[2] != 0x07) {
			num++;
			continue;
		}

		offset += 3;

		/* num_sta (2 byte) */
		if (offset + 2 > tlv_len)
			return false;

		num_sta = buf_get_be16(t + offset);
		offset += 2;

		for (i = 0; i < num_sta; i++) {
			/* macaddr(6 bytes) + tx_bytes(8 bytes)
			 * rx_bytes(8 bytes) + tx_packets(8 bytes)
			 * rx_packets(8 bytes)
			 */

			if (offset + 38 > tlv_len)
				return false;

			offset += 38;
		}
		num++;
	}

	return true;
}

bool validate_ap_autoconfig_response(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	struct tlv_policy a_policy[] = {
		[0] = { .type = TLV_TYPE_SUPPORTED_ROLE,
				.present = TLV_PRESENT_ONE,
				.minlen = 1, /* tlv_supported_role */
				.maxlen = 1,
		},
		[1] = { .type = MAP_TLV_SUPPORTED_SERVICE,
				.present = TLV_PRESENT_OPTIONAL_MORE,
				.minlen = 1, /* num of services */
		},
		[2] = { .type = MAP_TLV_MULTIAP_PROFILE,
				.present = TLV_PRESENT_OPTIONAL_MORE,
				.minlen = 1, /* tlv_map_profile */
				.maxlen = 1
		},
		[3] = { .type = TLV_TYPE_SUPPORTED_FREQ_BAND,
				.present = TLV_PRESENT_ONE,
				.minlen = 1, /* tlv_supported_band */
				.maxlen = 1
		}
	};
	int ret;

	dbg("%s |" MACFMT "|CMDU: ap autoconfig response\n",
		__func__, MAC2STR(cmdu->origin));

	ret = t_cmdu_parse_tlvs(cmdu, tv, a_policy, 4);
	if (ret) {
		dbg("%s: parse_tlv failed\n", __func__);
		return false;
	}
	if (!tv[0][0] || !tv[3][0]) {
		dbg("%s: Missing one or more mandatory TLV!\n", __func__);
		return false;
	}


	/* Parse SupportedRole TLV */
	if (check_supported_role_tlv(tv[0][0]))
		return false;

	/* Parse SupportedFreqBand TLV */
	if (check_supported_band_tlv(tv[3][0]))
		return false;

	return true;
}


bool validate_higher_layer_response(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{

	dbg("%s: --->\n", __func__);

	int ret = 0, num = 0;
	struct tlv_policy a_policy[] = {
		[0] = { .type = TLV_TYPE_AL_MAC_ADDRESS_TYPE,
			.present = TLV_PRESENT_ONE,
			.minlen = 6, /* tlv_almac_addr */
			.maxlen = 6,
		},
		[1] = { .type = TLV_TYPE_1905_PROFILE_VERSION,
			.present = TLV_PRESENT_ONE,
			.minlen = 1, /* tlv_profile_ver */
			.maxlen = 1,
		},
		[2] = { .type = TLV_TYPE_DEVICE_IDENTIFICATION,
			.present = TLV_PRESENT_ONE
		},
		[3] = { .type = TLV_TYPE_CONTROL_URL,
			.present = TLV_PRESENT_OPTIONAL_MORE
		},
		[4] = { .type = TLV_TYPE_IPV4,
			.present = TLV_PRESENT_OPTIONAL_MORE
		},
		[5] = { .type = TLV_TYPE_IPV6,
			.present = TLV_PRESENT_OPTIONAL_MORE
		},
	};

	dbg("parsing higher layer response |" MACFMT "|CMDU: higher layer response\n",
			MAC2STR(cmdu->origin));

	ret = t_cmdu_parse_tlvs(cmdu, tv, a_policy, 6);
	if (ret) {
		dbg("%s: parse_tlv failed\n", __func__);
		return false;
	}

	/* TLV_TYPE_AL_MAC_ADDRESS_TYPE */
	if (check_almac_addr_tlv(tv[0][0]))
		return false;

	/* TLV_TYPE_1905_PROFILE_VERSION */
	if (check_profile_ver_tlv(tv[1][0]))
		return false;

	/* TLV_TYPE_DEVICE_IDENTIFICATION */
	if (check_device_id_tlv(tv[2][0]))
		return false;

	/* TLV_TYPE_CONTROL_URL */
	if (tv[3][0]) {
		struct tlv_control_url *tlv;
		uint16_t tlv_len = 0;
		int size = 0;

		tlv_len = t_tlv_length(tv[3][0]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_control_url *)tv[3][0]->data;
		if (!tlv)
			return false;

		size = tlv_len - (sizeof(struct tlv_control_url));

		if (size != 0)
			return false;

	}

	num = 0;

	/* TLV_TYPE_IPV4 */
	while (tv[4][num]) {
		uint8_t *tv_data;
		struct tlv_ipv4 *tlv;
		uint16_t tlv_len = 0;
		int i = 0, offset = 0;
		int j = 0;

		tlv_len = t_tlv_length(tv[4][num]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_ipv4 *)tv[4][num]->data;
		if (!tlv)
			return false;

		tv_data = (uint8_t *)tlv;

		/*num_interfaces (1 byte)*/
		if (offset + 1 > tlv_len)
			return false;

		offset += 1;
		for (i = 0; i < tlv->num_interfaces; i++) {
			uint8_t num_ipv4 = 0;

			/* macaddr (6 bytes) */
			if (offset + 6 > tlv_len)
				return false;

			/*macaddr*/
			offset += 6;

			if (offset + 1 > tlv_len)
				return false;

			memcpy(&num_ipv4, &tv_data[offset], 1);

			/*num_ipv4 (1 byte)*/
			offset += 1;

			for (j = 0; j < num_ipv4; j++) {

				/* type (1 byte) + address (4 byte) + dhcpserver (4 byte) */
				if (offset + 9 > tlv_len)
					return false;

				/* type (1 byte) + address (4 byte) + dhcpserver (4 byte) */
				offset += 9;
			}
		}
		num++;
	}

	num = 0;
	/* TLV_TYPE_IPV6 */
	while (tv[5][num]) {
		uint8_t *tv_data;
		struct tlv_ipv6 *tlv;
		uint16_t tlv_len = 0;
		int i = 0, offset = 0;
		int j = 0;

		tlv_len = t_tlv_length(tv[5][num]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_ipv6 *)tv[5][num]->data;
		if (!tlv)
			return false;

		tv_data = (uint8_t *)tlv;

		/* num_interfaces (1 byte) */
		if (offset + 1 > tlv_len)
			return false;

		offset += 1;
		for (i = 0; i < tlv->num_interfaces; i++) {
			uint8_t num_ipv6 = 0;

			/* macaddr (6 bytes) */
			if (offset + 6 > tlv_len)
				return false;

			/* macaddr */
			offset += 6;

			/* link_local_address (16 bytes) */
			if (offset + 16 > tlv_len)
				return false;

			/* link_local_address */
			offset += 16;

			if (offset + 1 > tlv_len)
				return false;

			memcpy(&num_ipv6, &tv_data[offset], 1);

			/* num_ipv6 (1 byte) */
			offset += 1;

			for (j = 0; j < num_ipv6; j++) {

				/* type (1 byte) + address (16 byte) + origin (16 byte) */
				if (offset + 33 > tlv_len)
					return false;

				/* type (1 byte) + address (16 byte) + origin (16 byte) */
				offset += 33;
			}
		}
		num++;
	}
	return true;
}

bool validate_link_metric_response(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	 struct tlv_policy a_policy[] = {
		[0] = { .type = TLV_TYPE_TRANSMITTER_LINK_METRIC,
				.present = TLV_PRESENT_OPTIONAL_MORE,
		},
		[1] = { .type = TLV_TYPE_RECEIVER_LINK_METRIC,
				.present = TLV_PRESENT_OPTIONAL_MORE,
		}
	};
	int ret;
	int num = 0;

	dbg("%s |" MACFMT "|CMDU: link metric response\n",
		__func__, MAC2STR(cmdu->origin));

	ret = t_cmdu_parse_tlvs(cmdu, tv, a_policy, 2);
	if (ret) {
		dbg("%s: parse_tlv failed\n", __func__);
		return false;
	}

	/* TLV_TYPE_TRANSMITTER_LINK_METRIC */
	while (tv[0][num]) {
		struct tlv_tx_linkmetric *tlv;
		uint16_t tlv_len = 0;
		int offset = 0;
		int size = 0;

		tlv_len = t_tlv_length(tv[0][num]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_tx_linkmetric *)tv[0][num]->data;
		if (!tlv)
			return false;


		/* aladdr (6 byte) + neighbor aladdr (6 byte) */
		if (offset + 12 > tlv_len)
			return false;

		offset += 12;

		size = tlv_len - (sizeof(struct tlv_tx_linkmetric));

		if (size < 0)
			return false;
		else if (size > 0) {
			size = size % (sizeof(struct tx_link_info));
			if (size)
				return false;
		}
		num++;
	}

	num = 0;
	/* TLV_TYPE_RECEIVER_LINK_METRIC */
	while (tv[1][num]) {
		struct tlv_rx_linkmetric *tlv;
		uint16_t tlv_len = 0;
		int offset = 0;
		int size = 0;

		tlv_len = t_tlv_length(tv[1][num]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_rx_linkmetric *)tv[1][num]->data;
		if (!tlv)
			return false;


		/* aladdr (6 byte) + neighbor aladdr (6 byte) */
		if (offset + 12 > tlv_len)
			return false;

		offset += 12;

		size = tlv_len - (sizeof(struct tlv_rx_linkmetric));

		if (size < 0)
			return false;
		else if (size > 0) {
			size = size % (sizeof(struct rx_link_info));
			if (size)
				return false;
		}
		num++;
	}
	return true;
}
