/*
 * host_ieee1905_utils.c - host for multiap network ubus interface dependent on ieee1905
 *
 * Copyright (C) 2021 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: nevadita.chatterjee@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdbool.h>
#include <pthread.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <net/if.h>

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <uci.h>

#include <easy/easy.h>

#include <netlink/netlink.h>

#include "debug.h"
#include "host_config.h"
#include "topologyd.h"
#include "host_nodes.h"
#include "config.h"
#include "json_utils.h"
#include "host_utils.h"
#include "host.h"
#include "ieee1905_deps.h"
#include "host_ieee1905_utils.h"


void ieee1905_info_parser(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct topologyd_private *p = (struct topologyd_private *)req->priv;
	struct json_object *json_msg;
	char *json_str;

	json_str = blobmsg_format_json(msg, true);
	if (!json_str)
		return;

	json_msg = json_tokener_parse(json_str);
	if (!json_msg)
		goto out_str;

	if (p->ieee1905_interfaces) {
		json_object_put(p->ieee1905_interfaces);
		p->ieee1905_interfaces = NULL;
	}

	if (!json_object_is_type(json_msg, json_type_object))
		goto out_json;

	p->ieee1905_interfaces = json_object_get(json_msg);

out_json:
	json_object_put(json_msg);
out_str:
	free(json_str);
}

void host_get_ieee1905_info(struct topologyd_private *priv)
{
	struct blob_buf bb = {0};
	int ret;
	uint32_t id;
	int status;

	blob_buf_init(&bb, 0);

	dbg("Calling 'ieee1905 info'!!\n");

	if (!priv) {
		blob_buf_free(&bb);
		return;
	}

	status = ubus_lookup_id(priv->ctx, "ieee1905", &id);
	if (status != UBUS_STATUS_OK) {
		err("object 'ieee1905' not present!\n");
		blob_buf_free(&bb);
		return;
	}

	ret = ubus_invoke(priv->ctx, id, "info", bb.head,
			ieee1905_info_parser, priv, 2 * 1000);
	if (ret) {
		dbg("Failed to get ieee1905 info (ret = %d)\n", ret);
	}

	blob_buf_free(&bb);
}


void host_ieee1905_call(struct topologyd_private *priv)
{

	host_get_ieee1905_info(priv);
	if (priv->is_mapagent)
		host_send_topology_query_all(priv);
}

int topology_update_client_assoc_event(struct topologyd_private *priv, struct tlv_client_assoc_event *assoc)
{
	uint8_t hwaddr[6] = {0};
	uint8_t assoc_event = 0;

	dbg("|%s:%d| Inside\n", __func__, __LINE__);

	assoc_event = assoc->event;
	memcpy(hwaddr, assoc->macaddr, 6);

	dbg("Topology Notification TLV client addr " MACFMT " assoc_event 0x%02x\n",
		MAC2STR(hwaddr), assoc_event);

	if (assoc_event == 0x80) {
		struct host_node *hn;
		hn = host_node_lookup(priv->host.node_htable, hwaddr);
		if (!hn) {
			hn = host_topo_node_add(priv, NULL, hwaddr, 1);
			if (hn && hn->is_ipaddr)
				host_send_client_event(priv, hn, 1);
		} else {
			if (hn->active == 0) {
				hn = host_topo_node_add(priv, NULL, hwaddr, 1);
				if (hn && hn->is_ipaddr)
					host_send_client_event(priv, hn, 1);
			}
		}
	} else
		host_topo_node_del(priv, NULL, hwaddr, 1);
	return 0;
}

int topology_update_host_assoc_client(struct topologyd_private *priv,
				      struct tlv_assoc_client *tlv)
{
	struct host_node *hn = NULL;
	int i, offset = 0;
	uint8_t *pos = (uint8_t *) tlv;

	if (tlv == NULL || priv == NULL)
		return -1;

	offset += 1;
	dbg("TLV: assoc client\n");

	dbg("\tbss_nr: %d\n", tlv->num_bss);
	for (i = 0; i < tlv->num_bss; i++) {
		uint16_t num_client = 0;
		int j;

		dbg("\t\tbssid: " MACFMT "\n",
			MAC2STR(&pos[offset]));
		offset += 6;

		num_client = BUF_GET_BE16(pos[offset]);
		offset += 2;

		dbg("\t\tassoc_clients_nr: %u\n", num_client);
		for (j = 0; j < num_client; j++) {
			uint16_t conntime = 0;
			uint8_t client_addr[6];

			dbg("\t\t\tclient_addr is: " MACFMT "\n",
				MAC2STR(&pos[offset]));
			memcpy(client_addr, &pos[offset], 6);

			//Here look at the host node of this assoc client and update the time
			hn = host_node_lookup(priv->host.node_htable, &pos[offset]);
			if (!hn) {
				hn = host_topo_node_add(priv, NULL, &pos[offset], 1);
				if (hn && hn->is_ipaddr)
					host_send_client_event(priv, hn, 1);
			} else {
				if (hn->active == 0) {
					hn = host_topo_node_add(priv, NULL, &pos[offset], 1);
					if (hn && hn->is_ipaddr)
						host_send_client_event(priv, hn, 1);
				}
			}
			if (hn)
				hn->age_time = time(NULL);
			offset += 6;
			conntime = BUF_GET_BE16(pos[offset]);
			offset += 2;
			dbg("\t\t\tuptime: 0x%04x\n", conntime);
		}
	}
	return 0;
}

int topology_update_host_vendor_specific_data(struct node *n,
		struct topologyd_private *priv,
		struct tlv *t)
{
	struct host_node *hn = NULL;
	int i, offset = 0;
	int32_t sta_nr = 0;
	int ret = 0;
	struct tlv_vendor_metric_specific *tlv = NULL;

	dbg("|%s:%d| Inside\n", __func__, __LINE__);

	if (n == NULL || t == NULL || priv == NULL)
		return -1;

	if (t->len < MIN_VENDOR_TLV_LEN)
		return -1;

	/*Here we check that the vendor specific tlv should only be filled if its not selfnode*/
	ret = memcmp(n->hwaddr, priv->selfnode.hwaddr, HW_ADDR_LENGTH);
	if (ret == 0)
		return 0;

	tlv = (struct tlv_vendor_metric_specific *)t->data;
	if (tlv == NULL)
		return -1;

	if (!(tlv->oui[0] == IOPSYS_OUI_1 && tlv->oui[1] == IOPSYS_OUI_2 && tlv->oui[2] == IOPSYS_OUI_3)) {
		dbg("TLV: Vendor Specific tlv metric OUI not matching\n");
		return 0;
	}

	offset += 3;//oui
	sta_nr = BUF_GET_BE16(tlv->num_sta);
	offset += 2;
	dbg("TLV: Vendor Specific tlv metric\n");

	dbg("\tsta_nr: %d\n", sta_nr);
	for (i = 0; i < sta_nr; i++) {
		uint8_t client_addr[6];

		memcpy(client_addr, &t->data[offset], 6);
		dbg("\t\tmacaddr: " MACFMT "\n",
			MAC2STR(client_addr));

		//Here look at the host node of this assoc client and update the time
		hn = host_node_lookup(priv->host.node_htable, client_addr);
		offset += 6;
		if (!hn) {
			offset += 32;
			dbg("Node not found\n");
			continue;
		}
		/*offset increment by 32 as to skip Bytes for statistics
		statitstics collection is being done periodically in separate flow. */
		offset += 32;

		dbg("\t\ttxbytes: [%lld] rxbytes: [%lld]\n", hn->stats.tx_bytes, hn->stats.rx_bytes);
		dbg("\t\ttxpackets: [%lld] rxpackets: [%lld]\n", hn->stats.tx_packets, hn->stats.rx_packets);
	}
	return 0;
}


void host_send_topology_query_all(struct topologyd_private *priv)
{
	int i = 0;
	struct node *p;

	dbg("Inside %s %d\n", __func__, __LINE__);

	//Send the query to selfnode
	host_send_mid_topology_query(priv, NULL);
	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&priv->topo.node_htable[i]))
			continue;

		hlist_for_each_entry(p, &priv->topo.node_htable[i], hlist) {
			if (hwaddr_is_zero(p->hwaddr))  // TODO: unlikely..
				continue;
			host_send_mid_topology_query(priv, p->hwaddr);
		}
	}
}


struct cmdu_buff *host_topology_query_map_tlv(void)
{
	struct cmdu_buff *resp;
	uint16_t mid = 0;
	struct tlv *t;
	struct tlv_map_profile *data;
	int ret = 0;

	dbg("Inside %s...\n", __func__);
	resp = t_cmdu_alloc_simple(CMDU_TYPE_TOPOLOGY_QUERY, &mid);
	if (!resp) {
		err("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	dbg(" function %s mid = [%hu]\n", __func__, t_cmdu_get_mid(resp));
	t = t_cmdu_reserve_tlv(resp, 20);
	if (!t)
		goto out;

	t->type = MAP_TLV_MULTIAP_PROFILE;
	t->len = sizeof(*data);
	data = (struct tlv_map_profile *)t->data;
	data->profile = 0x02;

	ret = t_cmdu_put_tlv(resp, t);
	if (ret) {
		err("%s: error: cmdu_put_tlv()\n", __func__);
		goto out;
	}

	t_cmdu_put_eom(resp);
	return resp;
out:
	t_cmdu_free(resp);
	return NULL;
}

int host_send_mid_topology_query(struct topologyd_private *p, uint8_t *dst_mac)
{
	struct cmdu_buff *cmdu = NULL;

	cmdu = host_topology_query_map_tlv();
	if (cmdu == NULL)
		return -1;
	if (dst_mac == NULL) {
		send_cmdu(p, cmdu, p->selfnode.hwaddr);
	} else {
		send_cmdu(p, cmdu, dst_mac);
	}

	t_cmdu_free(cmdu);
	return 0;
}

