/*
 * topo_ieee1905.c - topology for ieee1905 multiap network
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdbool.h>
#include <errno.h>

#include <libubox/uloop.h>
#include <libubus.h>
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <uci.h>


#include <easy/easy.h>

#include "debug.h"
#include "host_config.h"
#include "topologyd.h"
#include "host.h"
#include "mdns_avahi.h"
#include "nodes.h"
#include "ieee1905_deps.h"

static bool first_run_flag;

static int topologyd_copy_update_node(struct node *dest, struct node *src)
{
	int j = 0, k = 0;

	if (dest == NULL || src == NULL)
		return -1;

	dest->nbr_num =  src->nbr_num;
	dest->non1905_nbr_num = src->non1905_nbr_num;

	//memset(dest->nbr_list, 0, sizeof(dest->nbr_list));
	memset(dest->non1905_nbrlist, 0, sizeof(dest->non1905_nbrlist));

	for (j = 0; j < src->nbr_num; j++)
		memcpy(dest->nbr_list[j].nbr_macid, src->nbr_list[j].nbr_macid, 6);

	memcpy(dest->non1905_nbrlist, src->non1905_nbrlist,
						6 * src->non1905_nbr_num);

	//copy the non1905 and 1905 neighbor local interface
	memcpy(dest->non1905_nbr_localintf, src->non1905_nbr_localintf, 6);
	memcpy(dest->nbr_localintf, src->nbr_localintf, 6);

	//Here copy the interface name of node
	strncpy(dest->ingress_ifr_name, src->ingress_ifr_name, strlen(src->ingress_ifr_name));

	//Here we copy the interface info
	dest->local_intf_nbr = src->local_intf_nbr;
	for (j = 0; j < src->local_intf_nbr; j++) {
		memcpy(dest->node_intf[j].interface_id, src->node_intf[j].interface_id, 6);
		dest->node_intf[j].media_type = src->node_intf[j].media_type;
		//Here we fill the power status data also
		dest->node_intf[j].power_status = src->node_intf[j].power_status;
		memset(dest->node_intf[j].generic_phy_oui, 0, sizeof(dest->node_intf[j].generic_phy_oui));
		dest->node_intf[j].generic_phy_variant = 0;
		memset(dest->node_intf[j].generic_phy_url, 0, sizeof(dest->node_intf[j].generic_phy_url));
		//Here we also need to fill the media specific
		//data for the interface
		dest->node_intf[j].media_specific_data_size = src->node_intf[j].media_specific_data_size;
		if (dest->node_intf[j].media_specific_data_size == 10) {
			memcpy(dest->node_intf[j].network_membership, src->node_intf[j].network_membership, 6);
			dest->node_intf[j].role = src->node_intf[j].role;
			dest->node_intf[j].ap_channel_band = src->node_intf[j].ap_channel_band;
			dest->node_intf[j].freq_index1 = src->node_intf[j].freq_index1;
			dest->node_intf[j].freq_index2 = src->node_intf[j].freq_index2;
		}
	}

	//Here we copy the L2 neighbor info
	dest->l2_nbr_num = src->l2_nbr_num;
	for (j = 0; j < src->l2_nbr_num; j++) {
		memcpy(dest->l2_nbr[j].l2_intf_addr, src->l2_nbr[j].l2_intf_addr, 6);
		memcpy(dest->l2_nbr[j].l2_nbr_addr, src->l2_nbr[j].l2_nbr_addr, 6);
		dest->l2_nbr[j].behind_intf_nbr = src->l2_nbr[j].behind_intf_nbr;
		for (k = 0; k < src->l2_nbr[j].behind_intf_nbr; k++)
			memcpy(dest->l2_nbr[j].behind_intf_addr[k], src->l2_nbr[j].behind_intf_addr[k], 6);

	}

	//Here we copy the Bridge tuple entries
	dest->bridge_tuple_num = src->bridge_tuple_num;
	for (j = 0; j < src->bridge_tuple_num; j++) {
		dest->br_tuple[j].tuple_num = src->br_tuple[j].tuple_num;

		for (k = 0; k < src->br_tuple[j].tuple_num ; k++)
			memcpy(dest->br_tuple[j].tuple_macid[k], src->br_tuple[j].tuple_macid[k], 6);
	}

	time(&dest->last_recv);

	return 0;
}

static void topologyd_node_expired(struct uloop_timeout *t)
{
	struct node *p = container_of(t, struct node, validity_timer);
	struct topologyd_private *priv = p->priv;
	int i = 0;
	struct topology_changelog elem;

	//Here before deleting a ieee1905 node
	//we also need to log in changelog the nonieee1905 nodes as deleted
	for (i = 0; i < p->non1905_nbr_num; i++) {
		dbg("Deleting the non1905 neighbor" MACFMT "\n", MAC2STR(p->non1905_nbrlist[i]));
		memcpy(elem.nbr_macaddr, p->non1905_nbrlist[i], 6);
		memcpy(elem.rpt_macaddr, p->hwaddr, 6);
		memcpy(elem.rpt_ifmacaddr, p->non1905_nbr_localintf, 6);
		elem.is1905_nbr = 0;
		elem.type = 1;
		enqueue_changelog(&(priv->topo), &elem);
		//Send event
		topology_send_node_event(&elem, p->ingress_ifr_name);
		/*Here we need to change the status of node in hosts*/
		host_topo_node_del(priv, NULL, elem.nbr_macaddr, 0);
	}

	//Here we log in the changelog as the node will be deleted
	changelog_copy_node_info(priv, p, 1, 1);

	if (priv) {
		node_del(priv->topo.node_htable, p->hwaddr);
		priv->topo.num_nodes--;
	}
}

int topologyd_node_add(struct topologyd_private *priv, struct node *node)
{
	struct node *p = NULL;
	bool newnode = 0;

	if (!node || !priv)
		return -1;

	if (priv->topo.num_nodes >= TOPOLOGY_NODE_MAX)
		return -1;

	if (!memcmp(node->hwaddr, priv->ieee1905_macaddr, 6)) {
		p = &priv->selfnode;

		if (first_run_flag == 1) {
			update_changelog_info(priv, p, node);
			update_ieee1905_node_info(priv, p, node);
		}

		topologyd_copy_update_node(p, node);
		if (first_run_flag == 0) {
			topologyd_update_changelog_firstentry(priv, p);
			/* Here we need to check the status of the
			 * hosts node that are read from the config
			 */
			host_run_status_check(priv);
			if (!priv->algo_running)
				topologyd_algo_run(priv);
			first_run_flag = 1;
		}
		return 0;
	}

	/* Add node if not present in topology.
	 *
	 * In case the node is present, update the node's values
	 * and the timestamp.
	 */
	p = node_lookup(priv->topo.node_htable, node->hwaddr);
	if (!p) {
		p = node_add(priv->topo.node_htable, node->hwaddr);
		topologyd_copy_update_node(p, node);

		p->validity_timer.cb = topologyd_node_expired;
		p->priv = priv;
		priv->topo.num_nodes++;
		//Add this in changelog events for ieee1905 nodes
		changelog_copy_node_info(priv, p, 0, 1);
		//Add this in changelog events for nonieee1905 nodes
		topologyd_update_changelog_firstentry(priv, p);
		newnode = 1;

	} else {
		//update in changelog
		update_changelog_info(priv, p, node);
		topologyd_copy_update_node(p, node);
	}

	/* reset validity timer for the node */
	uloop_timeout_set(&p->validity_timer, 75 * 1000);
	if ((!priv->algo_running) && newnode)
		topologyd_algo_run(priv);

	return 0;
}


/* Build full topology tree using a dumb but effective algo.
 * It uses ieee1905 topology query to probe the neighbor nodes.
 *
 * Assumption: all nodes keep up-to-date list of their immediate neighbors.
 *
 * At any node 'P' in the network -
 *
 * Intial state:
 *		Nbr(P) = {N0,..Nn}  <== immediate neighbors of P.
 *
 * Step 1:
 *		for all Ni in G:
 *		(a)	Send Topology Request to Ni neighbors
 *		(b)	Send Topology Request to Ni
 *
 * Step 2:
 *		Read Topology Response from Ni.
 *		if Nbr(Ni) is not in G:
 *			Insert Nbr(Ni) in G
 *
 * Step 3:
 *		Repeat Step 1-2 until no new Nbr(Ni) can be inserted in G.
 *
 */
void topologyd_algo_run(struct topologyd_private *priv)
{
	struct node *p = NULL, *node = NULL;
	int i, j;
	int is_selfnode = 0;
	//struct linkMetricQueryTLV tlv_metric;

	if (priv->algo_running)
		return;

	dbg("\n======Building Topology======\n");

	priv->algo_running = true;

	/* send query to own neighbors */
	p = &priv->selfnode;

	if (hwaddr_is_zero(p->hwaddr))
		topologyd_send_ieee1905_topology_query(priv, NULL);

	dbg("Processing " MACFMT " AL\n", MAC2STR(p->hwaddr));
	dbg("The total nodes are %d", p->nbr_num);

	for (j = 0; j < p->nbr_num; j++) {
		/*Here we need to see if the node exist in our hash table then req will 
		 * be sent from below */
		node = node_lookup(priv->topo.node_htable, p->nbr_list[j].nbr_macid);
		if (!node) {
			dbg("sending req to " MACFMT "\n", MAC2STR(p->nbr_list[j].nbr_macid));
			topologyd_send_ieee1905_topology_query(priv, p->nbr_list[j].nbr_macid);
			dbg("Processing " MACFMT " AL\n", MAC2STR(p->nbr_list[j].nbr_macid));
		}
	}
	/*here we send the topology query to selfnode each*/
	/*time the algo runs to keep self node updated */

	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&priv->topo.node_htable[i]))
			continue;

		hlist_for_each_entry(p, &priv->topo.node_htable[i], hlist) {

			if (hwaddr_is_zero(p->hwaddr))	// TODO: unlikely..
				continue;

			for (j = 0; j < p->nbr_num; j++) {
				node = NULL;
				is_selfnode = memcmp(&priv->selfnode, p->nbr_list[j].nbr_macid, 6);
				node = node_lookup(priv->topo.node_htable, p->nbr_list[j].nbr_macid);
				if (!node && is_selfnode) {
					dbg("sending req to " MACFMT "\n", MAC2STR(p->nbr_list[j].nbr_macid));
					topologyd_send_ieee1905_topology_query(priv, p->nbr_list[j].nbr_macid);
				}
			}
			//Here we also need to send the topology query to neighbors
			dbg("sending topology req to " MACFMT "\n", MAC2STR(p->hwaddr));
			topologyd_send_ieee1905_topology_query(priv, p->hwaddr);
			//Here we also need to send the query for higher layer request
			dbg("$$$$$$$$$$$$$$$$$$$$$$$$$sending higher layer req to " MACFMT "\n", MAC2STR(p->hwaddr));
			topologyd_send_ieee1905_query(priv, p->hwaddr, CMDU_TYPE_HIGHER_LAYER_QUERY, NULL);
			//Here sending the link metric query
			dbg("$$$$$$$$$$$$$$$$$$$$$$$$$sending linkmetric req to " MACFMT "\n", MAC2STR(p->hwaddr));
			//topologyd_getdata_lnmetric_query(&tlv_metric);
			topologyd_send_ieee1905_query(priv, p->hwaddr, CMDU_TYPE_LINK_METRIC_QUERY, NULL);
			//Here sending the generic phy query
		//	dbg("sending generic phy query req to " MACFMT "\n", MAC2STR(p->hwaddr));
		//	topologyd_send_ieee1905_query(priv, p->hwaddr, CMDU_TYPE_GENERIC_PHY_QUERY, p->ingress_ifr_name, NULL);

		}
	}

	priv->algo_running = false;

	host_remove_aged_node(priv);

	if (priv->status == 0)
		uloop_timeout_set(&priv->status_timer, TOPOLOGY_STATUS_INTERVAL*1000);

	/***
	 * gather service information through dns-sd
	 */
	mdns_init(priv);
	mdns_run(priv);
	mdns_cleanup(priv);

	dbg("\n====== Done Building Topology======\n");
}

int topologyd_map_sub_cb(void *bus, void *priv, void *data)
{
	struct blob_attr *msg = (struct blob_attr *)data;
	char *str;


	str = blobmsg_format_json(msg, true);
	dbg("Received notification '%s'\n", str);
	free(str);

	topologyd_event_handler(priv, msg);

	return 0;
}

int topologyd_map_del_cb(void *bus, void *priv, void *data)
{
	struct topologyd_private *p = (struct topologyd_private *)priv;
	uint32_t *obj = (uint32_t *)data;

	p->subscribed = false;
	fprintf(stdout, "Object 0x%x no longer present\n", *obj);

	return 0;
}

int topology_subscribe_for_cmdus(struct topologyd_private *p)
{
	mapmodule_cmdu_mask_t cmdu_mask = {0};
	uint32_t map_id;
	int ret;


	dbg("Inside  %s\n", __func__);
	t_map_prepare_cmdu_mask(cmdu_mask,
			CMDU_TYPE_TOPOLOGY_QUERY,
			CMDU_TYPE_HIGHER_LAYER_QUERY,
			CMDU_TYPE_LINK_METRIC_QUERY,
			CMDU_TYPE_TOPOLOGY_RESPONSE,
			CMDU_TYPE_HIGHER_LAYER_RESPONSE,
			CMDU_TYPE_TOPOLOGY_NOTIFICATION,
			CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE,
			CMDU_TYPE_LINK_METRIC_RESPONSE,
			CMDU_TYPE_TOPOLOGY_DISCOVERY,
			-1);


	dbg("<----------------------------------- %s\n", __func__);
	//dbg("cmdu-mask = 0x%x\n", i1905cmdus_mask);
	ret = ubus_lookup_id(p->ctx, map_plugin, &map_id);
	if (ret) {
		dbg("plugin '%s' lookup failed. %s\n",
			map_plugin, ubus_strerror(ret));
		return ret;
	}

	p->topo_oid = map_id;
	/* register as client to the map module */
	ret = t_map_subscribe(p->ctx,
			    &p->topo_oid,
			    "topologyd", &cmdu_mask, p,
			    topologyd_map_sub_cb,
			    topologyd_map_del_cb,
			    &p->subscriber);
	if (!ret) {
		p->subscribed = true;
	} else {
		dbg("Failed to 'register' with %s (err = %s)\n", map_plugin,
		    ubus_strerror(ret));
	}

	return ret;
}

void uobj_add_event_handler(struct ubus_context *ctx,
			    struct ubus_event_handler *ev,
			    const char *type, struct blob_attr *msg)
{

	char path[32] = {0};
	struct topologyd_private *priv = container_of(ev, struct topologyd_private, obj_register);
	struct blob_attr *tb[2];
	static const struct blobmsg_policy ev_attr[2] = {
		[0] = { .name = "id", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "path", .type = BLOBMSG_TYPE_STRING }
	};
	//dbg("msg - %s\n", blobmsg_format_json_indent(msg, true, -1));
	blobmsg_parse(ev_attr, 2, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1]) {
		dbg("%d %d\n", !tb[0], !tb[1]);
		return;
	}
	strncpy(path, blobmsg_data(tb[1]), sizeof(path) - 1);
	if (strncmp(path, map_plugin, strlen(map_plugin)))
		return;
	/* TODO: how to handle failure? */
	topology_subscribe_for_cmdus(priv);
}

